import sys
import scanpy as sc
import anndata
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt 
import matplotlib as mpl

import cell2location
import scvi

from matplotlib import rcParams
rcParams['pdf.fonttype'] = 42 # enables correct plotting of text
import seaborn as sns
from scipy import sparse

section = "P1"
reference_path = "./atlas/HGSOC_atlas.h5ad"
n_cell = 15
alpha= 200
n_epoch= 20000
use_gpu = True
section_path = "./Spatial/P1"

print('Done loading libraries and arguments')

def read_and_qc(sample_name, path):
    r""" 
    This function reads the data for one 10X spatial experiment into the anndata object.
    It also calculates QC metrics. Modify this function if required by your workflow.
    
    :param sample_name: Name of the sample
    :param path: path to data
    """
    import os
    import csv
    import scanpy as sc

    # Paths of the file
    file_path = path + '/spatial/tissue_positions.csv'
    new_file_path = path + '/spatial/tissue_positions_list.csv'

    # Check if the file exists
    if os.path.isfile(file_path):
        # Rename the file
        os.rename(file_path, new_file_path)
    
        # Remove the first line
        with open(new_file_path, 'r') as fin:
            data = fin.read().splitlines(True)
        with open(new_file_path, 'w') as fout:
            fout.writelines(data[1:])
        
        
    adata = sc.read_visium(path, count_file='filtered_feature_bc_matrix.h5', load_images=True, library_id='custom')
    adata.obs['sample'] = sample_name
    adata.var['SYMBOL'] = adata.var_names
    adata.var.rename(columns={'gene_ids': 'ENSEMBL'}, inplace=True)
    adata.var_names = adata.var['ENSEMBL']
    adata.var.drop(columns='ENSEMBL', inplace=True)

    # Calculate QC metrics
    sc.pp.calculate_qc_metrics(adata, inplace=True)
    adata.var['mt'] = [gene.startswith('mt-') for gene in adata.var['SYMBOL']]
    adata.obs['mt_frac'] = adata[:, adata.var['mt'].tolist()].X.sum(1).A.squeeze()/adata.obs['total_counts']

    # add sample name to obs names
    adata.obs["sample"] = [str(i) for i in adata.obs['sample']]
    adata.obs_names = adata.obs["sample"] + '_' + adata.obs_names
    adata.obs.index.name = 'spot_id'
    
    return adata



run_name = './'+section+'/cell2location_map'


#  Import visium data


adata_vis = read_and_qc(section, section_path)
print('Deconvolution: Done loading Visium section '+section)

# rename genes to ENSEMBL
adata_vis.var['gene_ids'] = adata_vis.var['SYMBOL']
adata_vis.var['SYMBOL'] = adata_vis.var_names
adata_vis.var_names = adata_vis.var['gene_ids']
adata_vis.var_names.name = None
adata_vis.var_names_make_unique()


slide = adata_vis



print('Deconvolution: Done processing Visium section '+section)
# find mitochondria-encoded (MT) genes
adata_vis.var['MT_gene'] = [gene.startswith('MT-') for gene in adata_vis.var['SYMBOL']]
# remove MT genes for spatial mapping (keeping their counts in the object)
adata_vis.obsm['MT'] = adata_vis[:, adata_vis.var['MT_gene'].values].X.toarray()
adata_vis = adata_vis[:, ~adata_vis.var['MT_gene'].values]



# LOAD scRNAseq reference

adata_file = reference_path
print('reference used for deconvolution:' + reference_path)
adata_ref = sc.read_h5ad(adata_file)
# export estimated expression in each cluster
if 'means_per_cluster_mu_fg' in adata_ref.varm.keys():
	inf_aver = adata_ref.varm['means_per_cluster_mu_fg'][[f'means_per_cluster_mu_fg_{i}'
                              for i in adata_ref.uns['mod']['factor_names']]].copy()
else:
	inf_aver = adata_ref.var[[f'means_per_cluster_mu_fg_{i}'
                              for i in adata_ref.uns['mod']['factor_names']]].copy()
inf_aver.columns = adata_ref.uns['mod']['factor_names']
print(inf_aver.iloc[0:5, 0:15])
#Cell2location: spatial mapping
# find shared genes and subset both anndata and reference signatures
intersect = np.intersect1d(adata_vis.var_names, inf_aver.index)
adata_vis = adata_vis[:, intersect].copy()
inf_aver = inf_aver.loc[intersect, :].copy()

# prepare anndata for cell2location model
cell2location.models.Cell2location.setup_anndata(adata=adata_vis, batch_key="sample")



# create and train the model
mod = cell2location.models.Cell2location(
adata_vis, cell_state_df=inf_aver,
# the expected average cell abundance: tissue-dependent
# hyper-prior which can be estimated from paired histology:
N_cells_per_location=int(n_cell),
# hyperparameter controlling normalisation of
# within-experiment variation in RNA detection (using default here):
detection_alpha=int(alpha)
)

mod.train(max_epochs=int(n_epoch),
     # train using full data (batch_size=None)
     batch_size=None,
     # use all data points in training because
     # we need to estimate cell abundance at all locations
     train_size=1,
     use_gpu=use_gpu)

#In this section, we export the estimated cell abundance (summary of the posterior distribution).
adata_vis = mod.export_posterior(
adata_vis, sample_kwargs={'num_samples': 1000, 'batch_size': mod.adata.n_obs, 'use_gpu': use_gpu}
)

# Save model
mod.save(f"{run_name}", overwrite=True)

# Save anndata object with results
adata_file = f"{run_name}/sp.h5ad"
adata_vis.write(adata_file)


